/*
** Module   : BufferImpl.java
** Abstract : implementation of Buffer interface
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description----------------------------------
** 001 ECF 20130118 Created initial version. Implementation of the Buffer
**                  interface; provides services for buffer-related, Progress
**                  methods, attributes, and some language statements.
** 002 VMN 20130207 Added support for handle-based methods
**                  BUFFER-COPY (KW_BUF_COPY)
**                  BUFFER-COMPARE (KW_BUF_COMP)
**                  BUFFER-FIELD (KW_BUF_FLD)
**                  BUFFER-RELEASE (KW_BUF_REL)
**                  FIND-BY-ROWID (KW_FIND_BR)
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
**                  and attributes
**                  NUM-FIELDS (KW_NUM_FLD)
**                  COLUMN-LABEL (KW_COL_LAB)
**                  INITIAL (KW_INIT)
**                  AMBIGUOUS (KW_AMBIG)
** 003 ECF 20130214 Added bufferCreate and variants of bufferCopy.
** 004 SVL 20130213 Added disable*Triggers functions.
** 005 VMN 20130217 Added support for handle-based attributes:
**                  CURRENT-ITERATION (KW_CUR_ITER)
**                  ERROR (KW_ERROR)
**                  DBNAME (KW_DBNAME)
**                  TABLE (KW_TABLE)
** 006 VMN 20130219 Added support for attributes
**                  TABLE-HANDLE (KW_TAB_HAND)
** 007 CA  20130221 Added support for ADM-DATA, UNIQUE-ID, NAMESPACE-URI. Implemented Deletable.
**                  Added missing BUFFER-FIELD stubs.
** 008 GES 20130224 Added support for a DELETE statement with a validation expression.
** 009 ECF 20130224 Added limited support for embedded SQL.
** 010 OM  20130222 Added bufferField variants and dereference from Dereferenceable.
** 011 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 012 VMN 20130306 Corrected support for handle-based methods:
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
** 013 VMN 20130311 Corrected support and runtime for handle-based methods:
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
**                  FIND-BY-ROWID (KW_FIND_BR)
**                  BUFFER-RELEASE (KW_BUF_REL)
**                  BUFFER-COMPARE (KW_BUF_COMP)
** 014 ECF 20130309 Changed API for bufferCopy and bufferCompare; only wrapped data types are
**                  allowed for literal parameters, to prevent the explosion of the API. 
** 015 VMN 20130314 Implemented runtime for handle-based methods:
**                  BUFFER-COMPARE (KW_BUF_COMP) also noLobs parameter type changed
**                  BUFFER-COPY (KW_BUF_COPY) (simple signatures)
**                  First parameter type changed from String to character for
**                  FIND-UNIQUE (KW_GET_UNI)
**                  FIND-FIRST (KW_FIND_1ST)
**                  FIND-LAST (KW_FIND_LST)
** 016 CA  20130323 Changed the APIs for KW_BUF_DEL and KW_EMPTY_TT to return logical.
** 017 ECF 20130417 Moved initial() method to FieldReference.
** 018 VMN 20130426 Implemented runtime for handle-based methods:
**                  BUFFER-COMPARE (KW_BUF_COMP)
**                  BUFFER-COPY (KW_BUF_COPY)
** 019 CA  20130603 resourceDelete must return a boolean value.
** 020 VMN 20130704 Implemented verifyTransaction() and added to handle-based methods:
**                  BUFFER-CREATE (KW_BUF_CREA)
**                  BUFFER-DELETE (KW_BUF_DELETE)
**                  FIND-BY-ROWID (KW_FIND_BR)
*/


package com.goldencode.p2j.persist;

import com.goldencode.p2j.util.*;

/**
 * This class implements services representing Progress builtin functions,
 * methods, attributes, and language statements, which operate on record
 * buffers.
 * <p>
 * Converted business logic invokes these methods on DMO proxies via a
 * DMO-specific interface in the form <code>&lt;DMO interface&gt;.Buf</code>.
 * Each such interface extends both a custom interface of DMO-specific getter
 * and setter methods, as well as the {@link Buffer} interface, which defines
 * the API for these standard services.
 * <p>
 * Dynamic proxy classes for each DMO type extend this class, so that business
 * logic can invoke these standard services directly as method calls on DMO
 * proxy instances.
 * <p>
 * This class is intentionally package private, since it is not meant to be
 * extended outside of this package (and within this package only by dynamic
 * proxy class), nor instantiated directly.
 * <p>
 * This class implements {@link BufferReference} in order to grant its
 * instance methods access to the {@link RecordBuffer} object associated with
 * each instance of the DMO proxy class which extends this class. This access
 * is necessary for these methods to do their work. Even though {@link
 * #buffer()} is a public method, external packages will not be able to invoke
 * this method (except through reflection). This is by design, as business
 * logic is not meant to have access to <code>RecordBuffer</code> instances.
 * The default implementation of this method here throws an exception, as it
 * is never intended to be called, but it needs to be present in order for
 * this class to compile, since it is referenced by other methods. Each DMO
 * proxy class overrides this method to provide a working implementation.
 */
public abstract class BufferImpl
extends HandleChain
implements Buffer
{
   /**
    * Worker static method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.findFirst() implementations
    *
    * @param   findQuery
    *          instance of FindQuery
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   private static logical wrapFirst(FindQuery findQuery, LockType lockType)
   {
      boolean success = true;
      try
      {
         findQuery.first(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Worker static method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.findLast() implementations
    *
    * @param   findQuery
    *          instance of FindQuery
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   private static logical wrapLast(FindQuery findQuery, LockType lockType)
   {
      boolean success = true;
      try
      {
         findQuery.last(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Worker static method for wrapping ErrorConditionException and reusing
    * in all P2JQuery.findUnique() implementations
    *
    * @param   findQuery
    *          instance of FindQuery
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   private static logical wrapUnique(FindQuery findQuery, LockType lockType)
   {
      boolean success = true;
      try
      {
         findQuery.unique(lockType);
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Checks if transaction is active otherwise records or shows error message.
    *
    * @return  <code>true</code> if a transaction is active else <code>false</code>.
    */
   private static boolean verifyTransaction()
   {
      if (TransactionManager.isTransaction())
      {
         return true;
      }
      else
      {
         String msg = "Unable to CREATE/DELETE unless a TRANSACTION is running";
         ErrorManager.recordOrShowError(7345, msg, false);
         return false;
      }
   }

   /**
   * Get the {@link RecordBuffer} instance associated with this object.
   * <p>
   * This implementation exists only to satisfy the Java compiler. This
   * method is only meaningful when a proxy DMO class is created by {@link
   * RecordBuffer}, which will implement this method to return an instance
   * of <code>RecordBuffer</code> associated with the proxy DMO instance.
   *
   * @return  Record buffer associated with this object.
   *
   * @throws  UnsupportedOperationException
   *          always.
   */
   protected abstract RecordBuffer buffer();
   
   /**
    * Reports if this object is valid for use.  
    *
    * @return  <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      // TODO:  implement correctly
      return true;
   }
   
   /**
    * Reports if this object is unknown.  
    *
    * @return  <code>true</code> if object is unknown.
    */
   public boolean unknown()
   {
      // TODO:  implement correctly
      return false;
   }
   
  /**
   * Get the record ID of the current buffer's current record, if any.  P2J
   * implementation of the RECID function.
   * <p>
   * This method will return a 32-bit recid which represents the internal
   * row ID of the given record.  If the actual row ID (which is implemented
   * as a wider integer) cannot fit due to overflow, an error condition will
   * be raised.
   *
   * @return  Record ID of the underlying data record. If no record currently
   *          backs the buffer associated with <code>dmo</code>, a recid
   *          initialized as unknown value is returned.
   * 
   * @throws  ErrorConditionException
   *          if the actual record ID is too wide to fit within a 32-bit
   *          integer.
   */
  public recid recordID()
  {
     return buffer().recordID();
  }
  
  /**
   * Get the row ID of the current buffer's current record, if any.  P2J
   * implementation of the ROWID function.
   *
   * @return  Row ID of the underlying data record.  If no record currently
   *          backs the buffer associated with <code>dmo</code>, a
   *          <code>rowid</code> initialized as unknown value is returned.
   */
  public rowid rowID()
  {
     return buffer().rowID();
  }
  
   /**
    * Indicate whether a backing data record currently is available.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   public logical available()
   {
      return (new logical(_available()));
   }
   
   /**
    * Indicate whether a backing data record currently is available.
    *
    * @return  <code>true</code> if data is available, <code>false</code> if
    *          no record backs this buffer.
    */
   public boolean _available()
   {
      return buffer().isAvailable();
   }
   
   /**
    * Checks if the record in the buffer has changed in database since the 
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   public logical currentChanged()
   {
      return new logical(_currentChanged());
   }
   
   /**
    * Checks if the record in the buffer has changed in database since the 
    * last <code>FIND CURRENT</code> or <code>GET CURRENT</code>.
    * P2J implementation of 4GL function <code>CURRENT-CHANGED</code>.
    *
    * @return  <code>true</code> if the database record differs from the
    *          value in memory buffer that was queried with the last
    *          <code>FIND</code> and <code>false</code> otherwise
    */
   public boolean _currentChanged()
   {
      return buffer().isRecordChanged();
   }
   
   /**
    * Indicate whether the buffer associated currently contains a record, and
    * whether that record is transient.  A record is transient if it has been
    * newly created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if a record exists and is new, else
    *          <code>false</code>.
    */
   public logical newlyCreated()
   {
      return (new logical(_newlyCreated()));
   }
   
   /**
    * Indicate whether the buffer associated currently contains a record, and
    * whether that record is transient.  A record is transient if it has been
    * newly created and has not yet been saved to the database.
    *
    * @return  <code>true</code> if a record exists and is new, else
    *          <code>false</code>.
    */
   public boolean _newlyCreated()
   {
      return buffer().isNewlyCreated();
   }
   
   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   public logical locked()
   {
      return (new logical(_locked()));
   }
   
   /**
    * Indicate whether the last record retrieval attempt was denied because
    * the target record was locked.
    *
    * @return  <code>true</code> if the record was locked;  <code>false</code>
    *          if the attempt succeeded, or if this query has not yet
    *          attempted to retrieve a record.
    */
   public boolean _locked()
   {
      return buffer().wasLocked();
   }
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   public logical ambiguous()
   {
      return (new logical(_ambiguous()));
   }
   
   /**
    * Indicate whether the last record retrieval attempt failed because the
    * query criteria was ambiguous.
    *
    * @return  <code>true</code> if the query was ambiguous;
    *          <code>false</code> if the attempt succeeded, or if no attempt
    *          has yet been made.
    */
   public boolean _ambiguous()
   {
      return buffer().wasAmbiguous();
   }
   
   /**
    * Create a new record in the buffer.
    *
    * @throws  ErrorConditionException
    *          if there is an error retrieving the next available primary key
    *          from the database or locking the record.
    */
   public void create()
   {
      buffer().create();
   }
   
   /**
    * Delete the record currently in the buffer, if any.
    *
    * @return  <code>true</code> if the delete was possible.
    *
    * @throws  ErrorConditionException
    *          if no record currently is loaded in the buffer, or if there is
    *          an error deleting the current record from the database.
    */
   public logical deleteRecord()
   {
      RecordBuffer buffer = buffer();

      if (!buffer.isTemporary() && !verifyTransaction())
      {
         return new logical(false);
      }
      try
      {
         buffer.delete();
         return new logical(true);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);

         return new logical(false);
      }
   }
   
   /**
    * Delete the record currently in the buffer, IF and ONLY IF the given
    * validator expression evaluates <code>true</code>.
    *
    * @param   validator
    *          The user-defined callback that is to be tested to determine if the
    *          delete should or should not occur.
    *
    * @throws  ErrorConditionException
    *          if no record currently is loaded in the buffer, or if there is
    *          an error deleting the current record from the database or if the
    *          validation expression fails.
    */
   public logical deleteRecord(BufferValidator validator)
   {
      // TODO: is this the correct behavior?
      if (validator.validate())
      {
         return deleteRecord();
      }
      else
      {
         return new logical(false);
      }
   }
   
   /**
    * Delete all records associated with this buffer, which match the criteria
    * specified by the given where clause.  Note that this where clause must
    * not contain DMO alias qualifiers.
    * <p>
    * <b>NOTE:</b>  the current implementation must not be invoked for a DMO
    * which is backed by a permanent table;  only temp-table bulk delete is
    * supported at this time.
    * 
    * @param   where
    *          An HQL where clause snippet which defines the restriction
    *          criteria to apply to the delete.  All references to properties
    *          in a DMO must be unqualified.
    * @param   args
    *          Query substitution parameters required by the where clause.
    * 
    * @throws  ErrorConditionException
    *          if there is an error deleting the records.
    * @throws  UnsupportedOperationException
    *          if invoked on a buffer associated with a permanent table.
    */
   public void delete(String where, Object ...args)
   {
      try
      {
         buffer().delete(where, args);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Bulk delete all records associated with the buffer.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the bulk delete operation.
    */
   public logical deleteAll()
   {
      buffer().deleteAll();
      
      // TODO: when does this return false?
      return new logical(true);
   }
   
   /**
    * Aggregate method which averages the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void average(String column, NumberType num)
   {
      // TODO:  implement
   }
   
   /**
    * Aggregate method which averages the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void average(String column, NumberType num, String where)
   {
      // TODO:  implement
   }
   
   /**
    * Aggregate method which averages the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void average(String column, NumberType num, String where, Object...args)
   {
      // TODO:  implement
   }
   
   /**
    * Aggregate method which counts the rows in the table backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate. May be the wildcard
    *          character ("*").
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void count(String column, NumberType num)
   {
      count(column, num, null);
   }
   
   /**
    * Aggregate method which counts the rows in the table backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate. May be the wildcard
    *          character ("*").
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void count(String column, NumberType num, String where)
   {
      count(column, num, where, null, (Object[]) null);
   }
   
   /**
    * Aggregate method which counts the rows in the table backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate. May be the wildcard
    *          character ("*").
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void count(String column, NumberType num, String where, Object...args)
   {
      buffer().count(column, num, where, args);
   }
   
   /**
    * Aggregate method which finds the minimum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void minimum(String column, NumberType num)
   {
      minimum(column, num, null);
   }
   
   /**
    * Aggregate method which finds the minimum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void minimum(String column, NumberType num, String where)
   {
      minimum(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which finds the minimum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void minimum(String column, NumberType num, String where, Object...args)
   {
      // TODO:  implement
   }
   
   /**
    * Aggregate method which finds the maximum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void maximum(String column, NumberType num)
   {
      maximum(column, null, null);
   }
   
   /**
    * Aggregate method which finds the maximum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void maximum(String column, NumberType num, String where)
   {
      maximum(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which finds the maximum value of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void maximum(String column, NumberType num, String where, Object...args)
   {
      // TODO:  implement
   }
   
   /**
    * Aggregate method which sums the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void sum(String column, NumberType num)
   {
      sum(column, num, null);
   }
   
   /**
    * Aggregate method which sums the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void sum(String column, NumberType num, String where)
   {
      sum(column, num, where, (Object[]) null);
   }
   
   /**
    * Aggregate method which sums the values of a single column across rows in the table
    * backing this buffer.
    * 
    * @param   column
    *          An expression indicating the name of the column to aggregate.
    * @param   num
    *          Variable into which to store the aggregate result.
    * @param   where
    *          HQL where clause used to filter the records.
    * @param   args
    *          Query substitution arguments.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the aggregate operation.
    */
   public void sum(String column, NumberType num, String where, Object...args)
   {
      // TODO:  implement
   }
   
   /**
    * Validate and release the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   public void release()
   {
      buffer().release();
   }
   
   /**
    * Validate the record currently in the buffer, if any.
    *
    * @throws  ErrorConditionException
    *          if the buffer fails validation.
    */
   public void validate()
   {
      buffer().validate();
   }
   
   /**
    * Open a read-only buffer scope (for trigger support).
    */
   public void openReadOnlyScope()
   {
      // TODO:  implement
   }

   /**
    * Conversion of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular buffer of a query or ProDataSet object.
    *
    * @param   fieldNumber
    *          field number. An INTEGER expression representing the sequence
    *          number of the field in the buffer.
    *
    * @return  handle to a particular field in the buffer.
    */
   @Override
   public handle bufferField(int fieldNumber)
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of BUFFER-FIELD() method (KW_BUF_FLD).
    *
    * Gets the handle to a particular buffer of a query or ProDataSet object.
    *
    * @param   fieldName
    *          field name. A CHARACTER string expression representing the name
    *          of the field in the buffer.
    *
    * @return  handle to a particular field in the buffer.
    */
   @Override
   public handle bufferField(String fieldName)
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   buffer
    *          The source buffer to copy from.
    *
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(Buffer buffer)
   {
      return bufferCopy(buffer, null, null);
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    *
    * @param   buffer
    *          The source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    *
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(Buffer buffer, character except)
   {
      return bufferCopy(buffer, except, null);
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   buffer
    *          The source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be copied.  Each pair must contain 
    *          one field name from the source and one field name from the target.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(Buffer buffer, character except, character pairs)
   {
      return RecordBuffer.copy(buffer(), ((BufferImpl)buffer).buffer(), except, pairs);
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   buffer
    *          The source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be copied.  Each pair must contain 
    *          one field name from the source and one field name from the target.
    * @param   noLobs
    *          If <code>TRUE</code>, BLOB and CLOB fields are ignored during the copy.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(Buffer buffer, character except, character pairs, logical noLobs)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   bufHandle
    *          The handle of the source buffer to copy from.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle)
   {
      return bufferCopy((Buffer) bufHandle.get());
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   bufHandle
    *          The handle of the source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle, character except)
   {
      return bufferCopy((Buffer) bufHandle.get(), except);
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   bufHandle
    *          The handle of the source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be copied.  Each pair must contain 
    *          one field name from the source and one field name from the target.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle, character except, character pairs)   {
      return bufferCopy((Buffer) bufHandle.get(), except, pairs);
   }

   /**
    * Conversion of BUFFER-COPY() method (KW_BUF_COPY).
    *
    * This method copies any common fields, determined by name, data type,
    * and extent-matching, from the source buffer to the receiving buffer.
    * If there are fields in one buffer that do not exist in the other, they
    * are ignored. This method is used to accommodate temp-tables of joins.
    * 
    * @param   bufHandle
    *          The handle of the source buffer to copy from.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the copy process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be copied.  Each pair must contain 
    *          one field name from the source and one field name from the target.
    * @param   noLobs
    *          If <code>TRUE</code>, BLOB and CLOB fields are ignored during the copy.
    * 
    * @return  <code>true</code> if BUFFER-COPY() succeeds, else <code>false</code>.
    */
   public logical bufferCopy(handle bufHandle,
                             character except,
                             character pairs,
                             logical noLobs)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   buffer
    *          Buffer with which to compare.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer buffer)
   {
      return bufferCompare(buffer, null, null, null);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   buffer
    *          Buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer buffer, character mode)
   {
      return bufferCompare(buffer, mode, null, null);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   buffer
    *          Buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer buffer, character mode, character except)
   {
      return bufferCompare(buffer, mode, except, null);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   srcBuffer
    *          Buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be compared.  Each pair must
    *          contain one field name from the source and one field name from the target.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer srcBuffer,
                                character mode,
                                character except,
                                character pairs)
   {
      return RecordBuffer.compare(((BufferImpl)srcBuffer).buffer(),
                                  buffer(),
                                  mode,
                                  except,
                                  pairs);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   buffer
    *          Buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be compared.  Each pair must
    *          contain one field name from the source and one field name from the target.
    * @param   noLobs
    *          If <code>TRUE</code>, BLOB and CLOB fields are ignored during the copy.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(Buffer    buffer,
                                character mode,
                                character except,
                                character pairs,
                                logical   noLobs)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle bufHandle)
   {
      return bufferCompare((Buffer) bufHandle.get());
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle bufHandle, character mode)
   {
      return bufferCompare((Buffer) bufHandle.get(), mode);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle bufHandle, character mode, character except)
   {
      return bufferCompare((Buffer) bufHandle.get(), mode, except);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be compared.  Each pair must
    *          contain one field name from the source and one field name from the target.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle bufHandle,
                                character mode,
                                character except,
                                character pairs)
   {
      return bufferCompare((Buffer) bufHandle.get(), mode, except, pairs);
   }

   /**
    * Conversion of BUFFER-COMPARE() method (KW_BUF_COMP).
    *
    * This method does a rough compare of any common fields, determined by
    * name, data type, and extent-matching, between the source buffer and the
    * target buffer. The resulting logical value is either TRUE or FALSE as
    * a whole. A single field that does not compare causes the entire buffer
    * to return FALSE. If there are fields in one buffer that do not exist in
    * the other, they are ignored.
    *
    * @param   bufHandle
    *          Handle of buffer with which to compare.
    * @param   mode
    *          The text "binary" or "case-sensitive", to provide that type of comparison.
    * @param   except
    *          A comma-separated list of fields that will be ignored in the compare process.
    * @param   pairs
    *          A comma-separated list of field-name pairs to be compared.  Each pair must
    *          contain one field name from the source and one field name from the target.
    * @param   noLobs
    *          If <code>TRUE</code>, BLOB and CLOB fields are ignored during the compare.
    *
    * @return  <code>true</code> if BUFFER-COMPARE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferCompare(handle    bufHandle,
                                character mode,
                                character except,
                                character pairs,
                                logical   noLobs)
   {
      throw new UnsupportedOperationException();
   }

   /**
    * Conversion of BUFFER-CREATE() method (KW_BUF_CREA).
    *
    * Creates a record, sets fields to their default values, and moves a copy of the record into 
    * the buffer.
    *
    * @return  <code>true</code> if successful
    */
   public logical bufferCreate()
   {
      if (!verifyTransaction())
      {
         return new logical(false);
      }
      boolean success = true;
      try
      {
         buffer().create();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Conversion of BUFFER-RELEASE() method (KW_BUF_REL).
    *
    * Releases a record from a buffer object. The BUFFER-RELEASE method
    * corresponds to the RELEASE statement.
    *
    * @return  <code>true</code> if BUFFER-RELEASE() succeeds,
    *          else <code>false</code>.
    */
   public logical bufferRelease()
   {
      boolean success = true;
      try
      {
         buffer().release();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Dereference a named member from this collection.
    *  
    * @param   memberName
    *          The name of the member to be extracted.
    *
    * @return  The value of the member or unknown (?) if there is no such member.
    */
   @Override
   public BaseDataType dereference(String memberName)
   {
      // TODO Auto-generated method stub
      return null;
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowID(rowid id)
   {
      return findByRowID(id, LockType.LT_SHARE);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the desired record.
    * @param   lockType
    *          Lock type
    *
    * @return  <code>true</code> if it finds the record, else <code>false</code>.
    */
   public logical findByRowID(rowid id, LockType lockType)
   {
      if (id.isUnknown())
      {
         String msg = "Invalid rowid for FIND-BY-ROWID method of BUFFER object " +
            buffer().getDMOAlias();
         ErrorManager.recordOrShowError(7341, msg, false);
         return new logical(false);
      }
      try
      {
         ErrorManager.silentErrorEnable();

         RecordBuffer buffer = buffer();
         buffer.initialize();

         // temporary, until support for dynamic query conversion at runtime to exactly match
         // Progress' behavior is not implemented
         String sort = null;
         FindQuery findQuery = new FindQuery(buffer(),
                                             buffer().getDMOAlias() + ".id = " + id.getValue(),
                                             null,
                                             sort,
                                             new Object[] { id },
                                             lockType);
         findQuery.unique();

         ErrorManager.silentErrorDisable();

         return new logical(!ErrorManager.isErrorFlag());
      }
      catch (IllegalArgumentException exc)
      {
      }
      catch (ErrorConditionException exc)
      {
      }
      catch (QueryOffEndException exc)
      {
      }
      return new logical(false);
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowID(rowid id, long lock)
   {
      return findByRowID(id, LockType.fromLockNowait(lock, 0));
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowID(rowid id, NumberType lock)
   {
      return findByRowID(id, LockType.fromLockNowait(lock.longValue(), 0));
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowID(rowid id, long lock, long nowait)
   {
      return findByRowID(id, LockType.fromLockNowait(lock, nowait));
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowID(rowid id, NumberType lock, long nowait)
   {
      return findByRowID(id, LockType.fromLockNowait(lock.longValue(), nowait));
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowID(rowid id, long lock, NumberType nowait)
   {
      return findByRowID(id, LockType.fromLockNowait(lock, nowait.longValue()));
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowID(rowid id, NumberType lock, NumberType nowait)
   {
      return findByRowID(id, LockType.fromLockNowait(lock.longValue(), nowait.longValue()));
   }

   /**
    * Conversion of FIND-BY-ROWID() method (KW_FIND_BR).
    *
    * Locates the record with the rowid you specify, then moves the record
    * into the buffer.
    *
    * @param   id
    *          An expression of type ROWID that represents the rowid of the
    *          desired record.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if it finds the record, else
    *          <code>false</code>.
    */
   public logical findByRowID(rowid id, integer lock, int nowait)
   {
      return findByRowID(id, LockType.fromLockNowait(lock.longValue(), nowait));
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst()
   {
      boolean success = true;
      try
      {
         new FindQuery(buffer(), null, null, null).first();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(character whereClause)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lockType
    *          Lock type
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(character whereClause, LockType lockType)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(character whereClause, long lock)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(character whereClause, NumberType lock)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(character whereClause, long lock, long nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(character whereClause, NumberType lock, long nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(character whereClause, long lock, NumberType nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-FIRST() method (KW_FIND_1ST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-FIRST() succeeds,
    *          else <code>false</code>.
    */
   public logical findFirst(character whereClause, NumberType lock, NumberType nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast()
   {
      boolean success = true;
      try
      {
         new FindQuery(buffer(), null, null, null).last();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(character whereClause)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lockType
    *          Lock type
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(character whereClause, LockType lockType)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(character whereClause, long lock)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(character whereClause, NumberType lock)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(character whereClause, long lock, long nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(character whereClause, NumberType lock, long nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(character whereClause, long lock, NumberType nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-LAST() method (KW_FIND_LST).
    *
    * Gets a single record. This method lets a user get the last record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-LAST() succeeds,
    *          else <code>false</code>.
    */
   public logical findLast(character whereClause, NumberType lock, NumberType nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique()
   {
      boolean success = true;
      try
      {
         new FindQuery(buffer(), null, null, null).unique();
      }
      catch (ErrorConditionException exc)
      {
         success = false;
      }
      return new logical(success);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(character whereClause)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lockType
    *          Lock type
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(character whereClause, LockType lockType)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(character whereClause, long lock)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(character whereClause, NumberType lock)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(character whereClause, long lock, long nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(character whereClause, NumberType lock, long nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(character whereClause, long lock, NumberType nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of FIND-UNIQUE() method (KW_FIND_UNI).
    *
    * Gets a single record. This method lets a user get a unique record that
    * satisfies the predicate expression.
    *
    * @param   whereClause
    *          where clause, predicate-expression.
    * @param   lock
    *          SHARE-LOCK|EXCLUSIVE-LOCK|NO-LOCK
    * @param   nowait
    *          NO-WAIT
    *
    * @return  <code>true</code> if FIND-UNIQUE() succeeds,
    *          else <code>false</code>.
    */
   public logical findUnique(character whereClause, NumberType lock, NumberType nowait)
   {
      // TODO:  implement
      return new logical(true);
   }

   /**
    * Conversion of NUM-FIELDS attribute (KW_NUM_FLD).
    *
    * Getter of NUM-FIELDS attribute.
    *
    * @return  current value of NUM-FIELDS attribute.
    */
   public integer numFields()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Disable the triggers associated with the FIND event for the buffer's table.
    */
   public void disableDumpTriggers()
   {
      // TODO:  implement
   }

   /**
    * Disable the triggers associated with the CREATE, WRITE, DELETE and ASSIGN events for the
    * buffer's table.
    *
    * @param allowReplication
    *        If <code>true</code> then only CREATE, DELETE, ASSIGN and WRITE triggers are
    *        disabled, and not REPLICATION-CREATE, REPLICATION-DELETE and REPLICATION-WRITE
    *        triggers.
    */
   public void disableLoadTriggers(boolean allowReplication)
   {
      // TODO:  implement
   }

   /**
    * Conversion of CURRENT-ITERATION attribute (KW_CUR_ITER).
    *
    * Getter of CURRENT-ITERATION attribute.
    *
    * @return current value of CURRENT-ITERATION attribute.
    */
   public integer currentIteration()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Returns the state of the ERROR attribute
    *
    * @return    Current state of the ERROR attribute.
    */
   public logical isError()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(boolean value)
   {
      // TODO:  implement
   }

   /**
    * Setter for ERROR attribute.
    *
    * @param    value
    *           The new value of ERROR attribute.
    */
   public void setError(logical value)
   {
      // TODO:  implement
   }

   /**
    * Returns the state of the DBNAME attribute
    *
    * @return    Current value of the DBNAME attribute.
    */
   public character getDbName()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Returns the state of the TABLE attribute
    *
    * @return    Current value of the TABLE attribute.
    */
   public character getTable()
   {
      // TODO:  implement
      return null;
   }

   /**
    * Conversion of TABLE-HANDLE attribute (KW_TAB_HAND).
    *
    * Gets the handle of a temp-table object, if any, associated with the buffer object.
    *
    * @return  the handle of a temp-table object.
    */
   public handle tableHandle()
   {
      // TODO:  implement
      return null;
   }
   
   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      // TODO Auto-generated method stub
      return null;
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(String value)
   {
      // TODO Auto-generated method stub
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      // TODO Auto-generated method stub
   }

   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself.
    * 
    * @return   The integer number of the children.
    */
   @Override
   public integer getUniqueID()
   {
      // TODO Auto-generated method stub
      return null;
   }

   /**
    * Returns the value of the the NAMESPACE-URI attribute.
    *
    * @return    See above.
    */
   @Override
   public character getNamespaceURI()
   {
      // TODO Auto-generated method stub
      return null;
   }

   /**
    * Setter for NAMESPACE-URI attribute.
    *
    * @param    uri
    *           The new value of the NAMESPACE-URI attribute.
    */
   @Override
   public void setNamespaceURI(String uri)
   {
      // TODO Auto-generated method stub
   }

   /**
    * Setter for NAMESPACE-URI attribute.
    *
    * @param    uri
    *           The new value of the NAMESPACE-URI attribute.
    */
   @Override
   public void setNamespaceURI(character uri)
   {
      // TODO Auto-generated method stub
   }

   @Override
   public handle bufferField(NumberType fieldNumber)
   {
      // TODO Auto-generated method stub
      return null;
   }

   @Override
   public handle bufferField(character fieldName)
   {
      // TODO Auto-generated method stub
      return null;
   }

   /**
    * Check if this resource supports the NEXT-SIBLING attribute.
    * 
    * @return   Always <code>false</code>
    */
   @Override
   protected boolean hasNextSibling()
   {
      return false;
   }

   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      // TODO Auto-generated method stub
      return false;
   }
}
